""" Detailed audit of incomplete test coverage reported by gitlab CI """

import ast
import http
import os
import zipfile
from pathlib import Path
from typing import List
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element

# pylint: disable=E0401, R0903, R1721
import requests

COVERAGE_REPORT_XML = "coverage.xml"


class CoverageReportSummary:
    """Wraps items parsed out of a coverage report"""

    def __init__(self):
        self.classes = self.functions = self.other_items = []

    def add_item(self, item):
        """
        Add this ast item to the appropriate list.

        :param item: ast thing, e.g., ClassDef, FunctionDef, If, IfExp
        :return:
        """
        if isinstance(item, ast.ClassDef):
            self.classes.append(item)
        elif isinstance(item, ast.FunctionDef):
            self.functions.append(item)
        else:
            self._add_other_item(item)

    def _add_other_item(self, item):
        """
        Examine this ast item and, if it's something we're interested in,
        add it to the other_items list.

        :param item: some ast thing that isn't a ClassDef nor a FunctionDef
        :return:
        """
        if isinstance(item, ast.If) or isinstance(item, ast.IfExp):
            self.other_items.append(item)
        elif not isinstance(item, ast.Import) and not isinstance(item, ast.ImportFrom):
            # (we're not interested in `import` lines)
            val = item.value
            if hasattr(val, "n"):
                content = item.value.n
            else:
                content = item.value
            if not isinstance(content, str) and not isinstance(
                    item, ast.Assign
            ):
                self.other_items.append(item)

class ModuleMetadata:
    """Represents a Python module to be audited for test coverage."""

    def __init__(self, path: Path, summary: CoverageReportSummary):
        self.path = path
        self.summary = summary


class ElementCoverageMetadata:
    """Encapsulates metadata from the coverage report for a module we're going to audit
    for test coverage"""

    def __init__(self, path: Path, line_rate: float, branch_rate: float):
        self.path = path
        self.line_rate = line_rate
        self.branch_rate = branch_rate
        self.modules = []

    def __str__(self):
        return f"{self.path}: line-rate = {self.line_rate}; branch-rate = {self.branch_rate}"

    def __eq__(self, other):
        if isinstance(other, ElementCoverageMetadata):
            return (
                other.path == self.path
                and other.line_rate == self.line_rate
                and other.branch_rate == self.branch_rate
                and other.modules == self.modules
            )
        return False

    def add_module(self, module: ModuleMetadata):
        """
        Add this module to the metadata.

        :param module: module to add
        :return:
        """
        self.modules.append(module)


class CoverageReportParser:
    """Turn a gitlab coverage report into audit targets we can examine"""

    def __init__(self, report_path: Path):
        self.xml_file = report_path
        self.project_root = find_project_root()
        self.targets = []

    def parse(self):
        """
        Find the audit targets in this coverage report

        :return: a list of metadata about classes to be audited
        """

        for target in self._read_coverage_report():
            self.targets.append(target)

    def _read_coverage_report(self) -> List[ElementCoverageMetadata]:
        """
        parse the coverage report generated by gitlab

        :return: a list of metadata about classes to be audited
        """

        targets = []
        coverage = ET.parse(self.xml_file)
        for package in coverage.findall(".//package"):
            complexity = int(package.attrib["complexity"])
            if complexity > 0:
                print(
                    f">>> (FYI: nonzero complexity of {complexity} for {package.attrib['name']})"
                )

            for element in package.findall(".//class"):
                target = ElementDigger(self.project_root, element).dig()
                if target:
                    targets.append(target)

        return targets


class ElementDigger:
    """Excavates an element parsed out of the coverage report"""

    def __init__(self, project_root: Path, element: Element):
        self.project_root = project_root
        self.element = element

    def dig(self) -> ElementCoverageMetadata:
        """
        Parse the class metadata for this "class" element of the coverage report.

        :return:
        """
        target = ElementCoverageMetadata(
            path=Path(self.element.attrib["filename"]),
            line_rate=float(self.element.attrib["line-rate"]),
            branch_rate=float(self.element.attrib["branch-rate"]),
        )

        add_target = False
        for lines in self.element.findall(".//lines"):
            for line in lines:
                attribs = line.attrib
                if "branch" in attribs and "condition-coverage" in attribs:
                    condition_coverage = attribs["condition-coverage"]
                    # it will be a % string; convert to an int
                    cov_percent = int(condition_coverage.split("%")[0])
                    if cov_percent < 100:
                        add_target = True
                        module = self._create_audit_module(target)
                        target.add_module(module)
        if add_target:
            return target

    def _create_audit_module(self, target: ElementCoverageMetadata) -> ModuleMetadata:
        """
        Find the Python module at this target's path, scoop up the functions and classes,
        and return an AuditModule

        :param target: a target of code test audit
        :return:
        """

        summary = CoverageReportSummary()

        module_path = self.project_root / target.path

        with open(module_path, "r") as infile:
            content = infile.read()
            dumped = ast.Dict(ast.parse(content))
            for item in dumped.keys.body:
                summary.add_item(item)

        return ModuleMetadata(path=module_path, summary=summary)


class CoverageReportGrabber:
    """Download gitlab's latest coverage report"""

    def __init__(self, destination: Path = None):
        if destination:
            self.destination = destination
        else:
            self.destination = Path.cwd()

    def grab(self):
        """
        Issue http GET request for coverage.xml, which will be contained n a .zip.
        Retrieve it to the current directory and unzip it.

        :return: the coverage report
        """

        url = "https://gitlab.nrao.edu/ssa/workspaces/-/jobs/artifacts/main/download?job=unit+test+coverage"
        response = requests.get(url)
        if response.status_code == http.HTTPStatus.OK:
            # write the zip file, then extract the .xml we want
            return self._write_cov_rpt_xml(response.content)

        raise RuntimeError(f"RETRIEVAL FAILURE: {response}")

    def _write_cov_rpt_xml(self, content: bytes) -> Path:
        """
        Write the coverage report extracted from the downloaded .zip

        :param content: file contents
        :return: COVERAGE_REPORT_XML file at self.destination
        """
        target_path = self.destination / "coverage.zip"
        with open(target_path, "wb") as cov_zip:
            cov_zip.write(content)

        xml_file = self.destination / COVERAGE_REPORT_XML
        with zipfile.ZipFile(target_path, "r") as zip_ref:
            zip_ref.extract(COVERAGE_REPORT_XML, self.destination)

        target_path.unlink()
        return xml_file


class Reporter:
    """Create a digest of not-fully-covered items found in the coverage report"""

    def __init__(self, targets: List[ElementCoverageMetadata]):
        self.targets = targets

    def report(self):
        """
        Print results to the console in human-readable format

        :return:
        """
        if len(self.targets) == 0:
            print("All workspaces modules are fully covered by current tests.")
            return

        print("The following workspaces modules have test coverage of < 100%:")
        for target in self.targets:
            print(target)


def we_are_in_docker() -> bool:
    """
    Are we executing inside a Docker container?
    (If not, we assume it's local testing.)

    :return: is she or isn't she? only her hairdresser knows for sure
    """
    return Path("/packages").is_dir()


def find_project_root() -> Path:
    """
    Under which top level do we start looking for tests?

    :return: project root
    """
    if we_are_in_docker():
        # we're testing in the workflow container
        return Path("/packages")

    # someone is testing locally
    possible_roots = [
        file for file in Path(os.environ["HOME"]).rglob("workspaces") if file.is_dir()
    ]
    if len(possible_roots) > 0:
        workspaces = possible_roots[0]
        return workspaces
    raise FileNotFoundError(f"project root not found under {os.environ['HOME' ]}")


def main(argv: List[str] = None):
    """
    yes, you -can- have it all

    :return:
    """
    if argv:
        target_path = Path(argv[0])
    else:
        target_path = Path.cwd()

    latest_report = CoverageReportGrabber(target_path).grab()
    parser = CoverageReportParser(latest_report)
    parser.parse()

    reporter = Reporter(parser.targets)
    reporter.report()


if __name__ == "main":
    main()
